seed=123

import numpy as np
np.random.seed(seed)
import tensorflow as tf
tf.set_random_seed(seed)

import random
random.seed(seed)

import os

import sys

import time

from my_bowl_dataset import BowlDataset

import model as modellib
from model import log

import numpy as np
from imgaug import augmenters as iaa

import skimage.io 



#######################################################################################
## SET UP CONFIGURATION
from config import Config

class BowlConfig(Config):
    """Configuration for training on the toy shapes dataset.
    Derives from the base Config class and overrides values specific
    to the toy shapes dataset.
    """
    # Give the configuration a recognizable name
    NAME = "003"

    IMAGE_RESIZE_MODE = "crop"
    # Augmentation parameters
    ASPECT_RATIO = 1.5 ## Maximum aspect ratio modification when scaling
    MIN_ENLARGE = 1.5 ## Minimum enlarging of images, note that this will be randomized
    ZOOM = 1.7 ## Maximum zoom per image, note that this will be randomized
    IMAGE_MIN_SCALE = False ## Not using this

    ROT_RANGE = 22.5
    CHANNEL_SHIFT_RANGE = 20
    
    LEARNING_RATE = 0.001
    # Train on 1 GPU and 8 images per GPU. We can put multiple images on each
    # GPU because the images are small. Batch size is 8 (GPUs * images/GPU).
    GPU_COUNT = 1
    IMAGES_PER_GPU = 2

    
    # Number of classes (including background)
    NUM_CLASSES = 1 + 1 # background + nuclei

    # Use small images for faster training. Set the limits of the small side
    # the large side, and that determines the image shape.
    IMAGE_MIN_DIM = 512
    IMAGE_MAX_DIM = 512

    # Use smaller anchors because our image and objects are small
    RPN_ANCHOR_SCALES = (8 , 16, 32, 64, 128)  # anchor side in pixels
    # RPN_ANCHOR_SCALES = (4, 8 , 16, 32, 64)  # anchor side in pixels

    # Reduce training ROIs per image because the images are small and have
    # few objects. Aim to allow ROI sampling to pick 33% positive ROIs.
    TRAIN_ROIS_PER_IMAGE = 600

    STEPS_PER_EPOCH = 664//IMAGES_PER_GPU
    VALIDATION_STEPS = 2//IMAGES_PER_GPU ## We are training with the whole dataset so validation is not very meaningfull, I put a two here so it is faster. We either use train loss or calculate in a separate procceses the mAP for each epoch

    # use small validation steps since the epoch is small
    # VALIDATION_STEPS = 5

    USE_MINI_MASK = True

    MAX_GT_INSTANCES = 256

    DETECTION_MAX_INSTANCES = 512

bowl_config = BowlConfig()
bowl_config.display()
#######################################################################################

# Root directory of the project
ROOT_DIR = os.getcwd()


## Change this dir to the stage 1 training data
train_dir = os.path.join(ROOT_DIR,'stage1_train')


# Get train IDs
train_ids = next(os.walk(train_dir))[1]

## Get an idea of img_sizes
train_paths = [os.path.join(train_dir, train_id,'images','{}.png'.format(train_id)) for train_id in train_ids]
# print(train_paths)
img_sizes=[]
for i in train_paths:
    image = skimage.io.imread(i)
    img_sizes.append(image.shape)

set(img_sizes)

## First pass to count instances
## same as:
# from collections import Counter
# Counter(img_sizes)

counts = []
for i in set(img_sizes):
    
    count = 0
    for j in np.arange(len(img_sizes)):
        if img_sizes[j] == i:
            # train_ids_size
            count = count +1
    print(i,count)
    counts.append(count)


long_train_ids = []
replic_n = max(counts)
for i in set(img_sizes):
    train_ids_size = []
    count = 0
    for j in np.arange(len(img_sizes)):
        if img_sizes[j] == i:
            train_ids_size.append(train_ids[j])
            # train_ids_size
            count = count +1
    repeticiones = int(replic_n/count)
    print(i,count,repeticiones)
    train_ids_size = np.repeat(train_ids_size,repeticiones)
    long_train_ids.extend(train_ids_size.tolist())



# train_ids = [os.path.join(train_dir, train_id) for train_id in train_ids]

train_ids = [os.path.join(train_dir, train_id) for train_id in long_train_ids]

# Training dataset
dataset_train = BowlDataset()
dataset_train.load_bowl(train_ids)
dataset_train.prepare()

# # Validation dataset, same as training.. will use pad64 on this one
dataset_val = BowlDataset()
dataset_val.load_bowl(train_ids)
dataset_val.prepare()






# Directory to save logs and trained model
MODEL_DIR = os.path.join(ROOT_DIR, "logs")

model = modellib.MaskRCNN(mode="training", config=bowl_config,
                          model_dir=MODEL_DIR)

## Change this with the path to the last epoch of train step 2
model_path = os.path.join(MODEL_DIR,'YOUR_LOG_DIR','final.h5') 
model.load_weights(model_path, by_name=True)

import time
start_time = time.time()

## Augment True will perform flipud fliplr and 90 degree rotations on the 512x512 images

# Image augmentation
# http://imgaug.readthedocs.io/en/latest/source/augmenters.html

# ## This should be the equivalente version of my augmentations using the imgaug library
# ## However, there are subtle differences so I keep my implementation
# augmentation = iaa.Sequential([
#     iaa.Fliplr(0.5),
#     iaa.Flipud(0.5),
#     iaa.OneOf([iaa.Affine(rotate=0),
#                iaa.Affine(rotate=90),
#                iaa.Affine(rotate=180),
#                iaa.Affine(rotate=270)]),
#     iaa.Sometimes(0.5,iaa.Affine(rotate=(-10,10))),
#     iaa.Add((-15, 15), per_channel=1)
    
# ])

augmentation=False

model.train(dataset_train, dataset_val, 
            learning_rate=bowl_config.LEARNING_RATE/30,
            epochs=5,
            augmentation=augmentation,
            layers="all")

end_time = time.time()
ellapsed_time = (end_time-start_time)/3600


print(model.log_dir)
model_path = os.path.join(model.log_dir, 'final.h5')
model.keras_model.save_weights(model_path)




import time
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
seed=123
from keras import backend as K

import numpy as np
np.random.seed(seed)
import tensorflow as tf

tf.set_random_seed(seed)

import random
random.seed(seed)

import skimage.io 
from skimage import img_as_ubyte
import skimage.transform

import model as modellib
import pandas as pd

import os


import my_functions as f


ROOT_DIR = os.getcwd()
MODEL_DIR = os.path.join(ROOT_DIR, "logs")

## This is the path to your trained weights
model_path = os.path.join(MODEL_DIR,'YOUR_LOG_PATH','final.h5') 

## modify these paths to where stage_2 images are downloaded
sample_submission = pd.read_csv('stage2_sample_submission_final.csv')
test_path ='stage_2' 
    

#######################################################################################
## SET UP CONFIGURATION
from config import Config

class BowlConfig(Config):
    """Configuration for training on the toy shapes dataset.
    Derives from the base Config class and overrides values specific
    to the toy shapes dataset.
    """
    # Give the configuration a recognizable name
    NAME = "Inference"

    IMAGE_RESIZE_MODE = "pad64" ## tried to modfied but I am using other git clone
    ## No augmentation
    ZOOM = False
    ASPECT_RATIO = 1
    MIN_ENLARGE = 1
    IMAGE_MIN_SCALE = False ## Not using this

    IMAGE_MIN_DIM = 512 # We scale small images up so that smallest side is 512
    IMAGE_MAX_DIM = False

    GPU_COUNT = 1
    IMAGES_PER_GPU = 1

    DETECTION_MAX_INSTANCES = 512
    DETECTION_NMS_THRESHOLD =  0.2
    DETECTION_MIN_CONFIDENCE = 0.9

    LEARNING_RATE = 0.001
    
    # Number of classes (including background)
    NUM_CLASSES = 1 + 1 # background + nuclei

    # Use smaller anchors because our image and objects are small
    RPN_ANCHOR_SCALES = (8 , 16, 32, 64, 128)  # anchor side in pixels

    # Reduce training ROIs per image because the images are small and have
    # few objects. Aim to allow ROI sampling to pick 33% positive ROIs.
    TRAIN_ROIS_PER_IMAGE = 600

    USE_MINI_MASK = True

inference_config = BowlConfig()
inference_config.display()
#######################################################################################










###################################################################################################
## Function that merges the test time augmentation masks for each image
## Note, this is not really model ensemble but test time augmentation (TTA) merge
###################################################################################################
def ensemble_image(aug_masks,image_id,iou_threshold = 0.6,remove_duplicate_threshold=0.7,threshold_rem=4,file_normal='foo.csv'):
    """ Function to merge the TTA, takes a list of n_augs TTAs and returns a single mask
    # Arguments
       aug_masks is a list of numpy arrays each containing a TTA prediction
       image_id string with image id
       iou_threshold: when two masks from two TTAs have an iou higher that this threshold they are considered the same particle
       remove_duplicate_threshold: value passed to the remove_duplicate function. Remove duplicates basically removes those pixels belonging to two different mask, in addition when doing this, if a mask area is reduced by a percentage that is more than this threshold it is removed from the mask
       threshold_rem: when a particle has not been detected by at least this number of TTA it is removed from the mask
    # Returns:
       The image_id and encoded pixels for that image
    """



    ##############################################################################################
    ##### MERGE TWO MASKS 
    ##############################################################################################
    def merge_masks(mask_1,mask_2,iou_threshold = 0.6,remove_duplicate_threshold=0.7):
        """ Function to merge two masks
        Code is based on the iou code that was provided as a function to calculate mAP
        # Arguments
           mask_1 and mask_2: the two numpy arrays with the masks to be merged
           iou_threshold: when two masks have an iou higher that this threshold they are considered the same particle
        # Returns:
           A single mask with the merged masks
        """

        ## clip mask_1 to 0 1 so we can flatten by addition
        ## preserve mask_1 which can have values larger than 1
        mask_1_0 = np.clip(mask_1,0,1)
        mask_1_0 = f.remove_duplicate(mask_1_0,threshold=remove_duplicate_threshold)
        mask_1 = mask_1[:,:,np.sum(mask_1_0,axis = (0,1)) != 0]
        mask_1_0 = mask_1_0[:,:,np.sum(mask_1_0,axis = (0,1)) != 0]

        pred_labels_1 = mask_1_0 * (np.arange(mask_1_0.shape[2])+1) ## replace 1 by the number of mask
        pred_labels_1= np.sum(pred_labels_1,axis=2) ## flatten by addition

        ## flatten the second mask, this one has only 0 and 1s
        mask_2 = f.remove_duplicate(mask_2,threshold=remove_duplicate_threshold)
        pred_labels_2 = mask_2 * (np.arange(mask_2.shape[2])+1) ## replace 1 by the number of mask
        pred_labels_2= np.sum(pred_labels_2,axis=2) ## flatten by addition

        ## Number of objects for mask1 and mask2
        true_objects = len(np.unique(pred_labels_1))
        pred_objects = len(np.unique(pred_labels_2))

        # Compute intersection between all objects
        intersection = np.histogram2d(pred_labels_1.flatten(), pred_labels_2.flatten(), bins=(true_objects, pred_objects))[0]

        # Compute areas (needed for finding the union between all objects)
        area_true = np.histogram(pred_labels_1, bins = true_objects)[0]
        area_pred = np.histogram(pred_labels_2, bins = pred_objects)[0]
        area_true = np.expand_dims(area_true, -1)
        area_pred = np.expand_dims(area_pred, 0)

        # Compute union
        union = area_true + area_pred - intersection

        # Compute the intersection over union
        iou = intersection / union
        best_match =(np.argmax(iou,axis=1)-1)[1:]
        best_match_iou =(np.max(iou,axis=1))[1:]
        best_match[best_match_iou<iou_threshold] = -1

        ## remove zeros as it is either background or no match
        second_labels= np.arange(pred_objects-1) ## background is not an object
        second_not_one = np.setdiff1d(second_labels,best_match)
        mask_4 = mask_2[:,:,second_not_one]  ## Masks in 2 not in 1, we divide by two

        ##  Append a zero mask at the end of  mask_2 that will be used for no matches
        zeromat= np.expand_dims(np.zeros_like(mask_2[:,:,0]),2)
        mask_3 = np.concatenate((mask_2,zeromat) ,axis = 2)
        mask_3 = mask_3[:,:,best_match] ## ordered mask_2 elements

        ## Compose the averaged masks
        merged_mask = np.sum(np.array([mask_1,mask_3]), axis=0)
        merged_mask = np.concatenate((merged_mask,mask_4) ,axis = 2)
        return merged_mask
    ##############################################################################################


    
    ## Default empty values in case no mask found
    ImageId_i =  [image_id]
    EncodedPixels_i = ['']
    
    mask_1 = aug_masks[0]
    for j in np.arange(1,len(aug_masks)):
        mask_2 = aug_masks[j]
        if (type(mask_1)== bool): ## Mask1 had no objects copy mask_2 over
            print('No valid mask_1, skip')
            mask_1 = mask_2
        else:
            if (type(mask_2)!= bool): ## Both masks have objects
                print('Merging',j,' for image' , image_id)
                mask_1 = merge_masks(mask_1,mask_2,iou_threshold = 0.6,remove_duplicate_threshold= remove_duplicate_threshold)
            else:
                print('No Merge, keep mask_1 as is')
    if (type(mask_1)!= bool): # otherwise after merging we still have no particles detected
        ## Remove pixels that were not detected by at least threshold_rem TTA and empty masks
        mask_1 = (mask_1>threshold_rem).astype('uint8') 
        _idx = np.sum(mask_1, axis=(0, 1)) > 0
        mask_1 = mask_1[:, :, _idx]


        if(mask_1.shape[2]>0):  ## Image has no detected nuclei
            ImageId_batch, EncodedPixels_batch, mask_1 = f.numpy2encoding(mask_1, image_id)
            ImageId_i = ImageId_batch
            EncodedPixels_i = EncodedPixels_batch

            ImageId_batch, EncodedPixels_batch, mask_1 = f.numpy2encoding(mask_1, image_id,dilation=True)
    lock.acquire()
    df = pd.DataFrame({ 'ImageId' : ImageId_i , 'EncodedPixels' : EncodedPixels_i})
    df.to_csv(file_normal, index=False,mode='a', columns=['ImageId', 'EncodedPixels'],header= False)

    lock.release()
    
    return 0
###################################################################################################





###################################################################################################
## Define a single thread queue
## that will be merging the image masks while the TTA continues on other images
###################################################################################################

def worker(queue) :
    while True :
        items = queue.get(True)
        ensemble_image(items[0],items[1],iou_threshold = items[2],remove_duplicate_threshold=items[3],threshold_rem=items[4],file_normal=items[5])

import multiprocessing
lock = multiprocessing.Lock()
###################################################################################################




# Recreate the model in inference mode
model = modellib.MaskRCNN(mode="inference", 
                          config=inference_config,
                          model_dir=MODEL_DIR)

print("Loading weights from ", model_path)
model.load_weights(model_path, by_name=True)


## We will predict on 15 different augmentations:
n_augs = 15
# FlipUD || FlipLR || ROT90  || CHANNEL_SHIFT_RANGE || ENLARGE_W || ENLARGE_H
aug_options = [[False, False, 0, 0,1,1],
               [False, False, 0, 0,2,2],
               [False, False, 0, 0,0.7,0.8],
               [False, False, 0, 15,1.5,1.5],
               [True, True, 1,0,0.9,0.8],
               [True, False, 2,7,1.1,1.2],
               [False,True, 3,10,0.5,0.5],
               [False,True, 3,10,1,1.2],
               [True,True, 2,25,1,1],
               [True,True, 2,25,1,1],
               [False,True, 3,5,1,1],
               [True,False, 1,15,1,1],
               [True,True, 3,15,1,0.9],
               [False,True, 3,10,1.5,1.5],
               [False, False, 3,5,1.2,1]]




ImageId = []
EncodedPixels = []

start=time.time()

iou_threshold = 0.6
remove_duplicate_threshold=0.7
threshold_rem=4

file_normal='submission_tta.csv'
## Write the headers of the submission files
df = pd.DataFrame({ 'ImageId' : [] , 'EncodedPixels' : []})
df.to_csv(file_normal, index=False, columns=['ImageId', 'EncodedPixels'])



## Multiproccesing job list
jobs=[]

n_images= len(sample_submission.ImageId)
for i in np.arange(n_images):
    image_id = sample_submission.ImageId[i]
    ##Set seeds for each image
    print('Start detect',i, '  ' ,image_id)
    random.seed(seed)
    np.random.seed(seed)
    tf.set_random_seed(seed)

    image_path = os.path.join(test_path, image_id, 'images', image_id + '.png')
    original_image = skimage.io.imread(image_path)
    ####################################################################
    ## This is needed for the stage 2 image that has only one channel
    if len(original_image.shape)<3:
        original_image = img_as_ubyte(original_image)
        original_image = np.expand_dims(original_image,2)
        original_image = original_image[:,:,[0,0,0]] # flip r and b
    ####################################################################

    original_image = original_image[:,:,:3]

    bzResults = [] 
    ## Predict for each augmentation
    for j in np.arange(n_augs): ## for each augmentation
        print('Model', j, 'Image',i, 'id ', image_id)
        flipud, fliplr, nrot, shift_range, h_ratio, w_ratio = aug_options[j]
        ## IMAGE AS DIRECTLY READ FROM DISK
        image = np.copy(original_image)
        if flipud:
            image = np.flipud(image)
        if fliplr:
            image = np.fliplr(image)
        if nrot:
            image = np.rot90(image, k=nrot,axes=(0,1))
        # if shift_range:
        #     image = random_channel_shift(image,shift_range,2)
        h, w = image.shape[:2] ## has to be after rotations
        if h_ratio !=1 or w_ratio!= 1:
            if ((h <512 or w < 512) and (h_ratio<1 or w_ratio<1)):##
                print('skip reduce, image too small',image.shape,h_ratio,w_ratio)
            else:
                scale = (h_ratio, w_ratio) ## change aspect ratio and enlarge
                round_scale = (round(h * scale[0]), round(w * scale[1]))
                image = skimage.transform.resize(
                    image, (round(h * scale[0]), round(w * scale[1])),
                    order=1, mode="constant", preserve_range=True)


        ## Make a prediction for this image and augmentation
        results = model.detect([image], verbose=0)
        r=results[0]
        mask = results[0]['masks']
        print('End detection')

        ## If no particle detected the code returns masks as minimasks
        ## convert to False
        if len(r['class_ids']):

            ## Now we have to reverse the image augmentations so we get back the original image
            h_2,w_2 = mask.shape[:2] ## has to be before rotations, reverse order augmentation
            if h_ratio !=1 or w_ratio!= 1:
                if ((h <512 or w < 512) and (h_ratio<1 or w_ratio<1)):##
                    print('skip reduce, image too small',image.shape,h_ratio,w_ratio)
                else:
                    mask = scipy.ndimage.zoom(mask, zoom=[h/h_2, w/w_2,  1], order=0)
            if nrot:
                mask = np.rot90(mask, k=4-nrot,axes=(0,1))
            if fliplr:
                mask = np.fliplr(mask)
            if flipud:
                mask = np.flipud(mask)
            if not np.all(mask.shape[:2] == original_image.shape[:2]):
                print ('Mask shape', mask.shape, 'and image shapes', original_image.shape, ' differ')
                break

        else:
            mask = False

        print('End ')

        bzResults.append(mask) ## append mask to the list of augmented masks for that image

    p= multiprocessing.Process(target=ensemble_image,args=(bzResults,image_id,iou_threshold,remove_duplicate_threshold,threshold_rem,file_normal,))
    jobs.append(p)
    p.start()

    #######################################################################
    ## Do not overpopulate the multiproccessing jobs
    MAXPROCCESES = 10
    wait_for_proccesses=True
    while wait_for_proccesses:
        for proc_i in np.flip(np.arange(len(jobs)),0): ## Need reverse ordering so we can delete without changing indexing
            proc= jobs[proc_i]
            proc.join(timeout=0)
            if not proc.is_alive(): ## Proccess has finished, will remove it from list
                del jobs[proc_i]
            
        if len(jobs)<MAXPROCCESES: 
            wait_for_proccesses = False
        print("Alive Jobs",len(jobs))


## Wait for all proccesses to finish
for j in jobs:
    j.join()
    
end = time.time()




import os
import shutil

from mrcnn.my_inference import predict_images
from mrcnn.preprocess_images import preprocess_images
from mrcnn.convert_to_image import convert_to_image, convert_to_imagej

def perform_segmentation(input_dir, output_dir):
  """
  Performs image segmentation tasks with predefined options.

  Args:
      input_dir (str): Path to the directory containing images for segmentation.
      output_dir (str): Path to the directory for saving outputs.
  """

  # Define segmentation options
  rescale = False
  scale_factor = 2  # Factor to downsize images by if rescale is True
  save_preprocessed = True
  save_compressed = True
  save_masks = True
  verbose = True
  output_imagej = True

  # Create output directory if it doesn't exist
  if output_dir != '' and not os.path.isdir(output_dir):
    os.mkdir(output_dir)

  # Check if output directory is empty
  if os.path.isdir(output_dir) and len(os.listdir(output_dir)) > 0:
    print("ERROR: Make sure the output directory is empty.")
    return

  # Define subdirectory paths based on output_dir
  preprocessed_image_dir = os.path.join(output_dir, "preprocessed_images/")
  preprocessed_image_list = os.path.join(output_dir, "preprocessed_images_list.csv")
  rle_file = os.path.join(output_dir, "compressed_masks.csv")
  output_mask_dir = os.path.join(output_dir, "masks/")
  output_imagej_dir = os.path.join(output_dir, "imagej/")

  # Preprocess the images
  if verbose:
    print("\nPreprocessing your images...")
  preprocess_images(input_dir,
                    preprocessed_image_dir,
                    preprocessed_image_list,
                    verbose=verbose)

  # Run inference on the neural network
  if verbose:
    print("\nRunning your images through the neural network...")
  predict_images(preprocessed_image_dir,
                  preprocessed_image_list,
                  rle_file,
                  rescale=rescale,
                  scale_factor=scale_factor,
                  verbose=verbose)

  # Save masks based on options
  if save_masks:
    if verbose:
      print("\nSaving the masks...")

    if output_imagej:
      convert_to_image(rle_file,
                       output_mask_dir,
                       preprocessed_image_list,
                       rescale=rescale,
                       scale_factor=scale_factor,
                       verbose=verbose)

      convert_to_imagej(output_mask_dir,
                       output_imagej_dir)
    else:
      convert_to_image(rle_file,
                       output_mask_dir,
                       preprocessed_image_list,
                       rescale=rescale,
                       scale_factor=scale_factor,
                       verbose=verbose)

  # Cleanup temporary files
  os.remove(preprocessed_image_list)

  if not save_preprocessed:
    shutil.rmtree(preprocessed_image_dir)

  if not save_compressed:
    os.remove(rle_file)

  if not save_masks:
    shutil.rmtree(output_mask_dir)

# Example usage
if __name__ == "__main__":
  perform_segmentation("/16Tbdrive1/vishakhag/CamAge_Package_ingredients/sample1/",
                      "/16Tbdrive1/vishakhag/CamAge_Package_ingredients/seg_outputs/out4/")



import numpy as np
from PIL import Image
import os
import skimage.exposure

'''Convert input images to RGB format in separate folders required by MRCNN

Arguments:
inputdirectory: Input directory containing images.
outputdirectory: Output directory to put new files in.
outputfile: Path to save comma-delimited file that will tell the neural network the image paths.'''
def preprocess_images(inputdirectory, outputdirectory, outputfile, verbose = False):
    if inputdirectory[-1] != "/":
        inputdirectory = inputdirectory + "/"
    if outputdirectory[-1] != "/":
        outputdirectory = outputdirectory + "/"

    if not os.path.exists(outputdirectory):
        os.makedirs(outputdirectory)

    output = open(outputfile, "w")
    output.write("ImageId, EncodedRLE" + "\n")
    output.close()

    for imagename in os.listdir(inputdirectory):
        try:
            if verbose:
                print ("Preprocessing ", imagename)

            image = np.array(Image.open(inputdirectory + imagename))
            if len(image.shape) > 2:
                image = image[:, :, 0]
            height = image.shape[0]
            width = image.shape[1]

            # Preprocessing operations
            image = skimage.exposure.rescale_intensity(image.astype(np.float32), out_range=(0, 1))
            image = np.round(image * 255).astype(np.uint8)        #convert to 8 bit
            image = np.expand_dims(image, axis=-1)
            rgbimage = np.tile(image, 3)                          #convert to RGB

            imagename = imagename.split(".")[0]

            if not os.path.exists(outputdirectory + imagename):
                os.makedirs(outputdirectory + imagename)
                os.makedirs(outputdirectory + imagename + "/images/")
            rgbimage = Image.fromarray(rgbimage)
            rgbimage.save(outputdirectory + imagename + "/images/" + imagename + ".png")

            output = open(outputfile, "a")
            output.write(imagename + ", " + str(height) + " " + str(width) + "\n")
            output.close()
        except IOError:
            pass






import numpy as np
import csv
from PIL import Image
import skimage.transform
import time
import os
import cv2
'''Converts the compression rle files to images.

Input:
rlefile: csv file containing compressed masks from the segmentation algorithm
outputdirectory: directory to write images to
preprocessed_image_list: csv file containing list of images and their heights and widths'''
def convert_to_image(rlefile, outputdirectory, preprocessed_image_list,
                     rescale = False, scale_factor = 2, verbose = False):

    rle = csv.reader(open(rlefile), delimiter=',')
    rle = np.array([row for row in rle])[1:, :]

    image_list = csv.reader(open(preprocessed_image_list), delimiter=',')
    image_list = np.array([row for row in image_list])[1:, :]

    def rleToMask(rleString,height,width):
      rows,cols = height,width
      rleNumbers = [int(numstring) for numstring in rleString.split(' ')]
      rlePairs = np.array(rleNumbers).reshape(-1,2)
      img = np.zeros(rows*cols,dtype=np.uint8)
      for index,length in rlePairs:
        index -= 1
        img[index:index+length] = 255
      img = img.reshape(cols,rows)
      img = img.T
      return img

    if outputdirectory[-1] != "/":
        outputdirectory = outputdirectory + "/"
    if not os.path.exists(outputdirectory):
        os.makedirs(outputdirectory)

    files = np.unique(rle[:, 0])
    for f in files:
        if verbose:
            start_time = time.time()
            print ("Converting", f, "to mask...")

        list_index = np.where(image_list[:, 0] == f)[0][0]
        file_string = image_list[list_index, 1]

        size = file_string.split(" ")
        height = np.int(size[1])
        width = np.int(size[2])

        new_height = height
        new_width = width
        if rescale:
            new_height = height // scale_factor
            new_width = width // scale_factor

        image = np.zeros((new_height, new_width)).astype(np.float32)
        columns = np.where(rle[:, 0] == f)
        currobj = 1
        for i in columns[0]:
            currimg = rleToMask(rle[i, 1], new_height, new_width)
            currimg = currimg > 1
            image = image + (currimg * currobj)
            currobj = currobj + 1

        if rescale:
            image = skimage.transform.resize(image, output_shape = (height, width), order=0, preserve_range = True)
        cv2.imwrite(outputdirectory + f + ".jpg",image)
        image = Image.fromarray(image)
        image.save(outputdirectory + f + ".tif")
 


        if verbose:
            print ("Completed in", time.time() - start_time)
            
def convert_to_imagej(inputdirectory, outputdirectory):
    import os
    import numpy as np
    from PIL import Image
    import skimage.segmentation as seg

    if outputdirectory[-1] != "/":
        outputdirectory = outputdirectory + "/"
    if not os.path.exists(outputdirectory):
        os.makedirs(outputdirectory)

    for filename in os.listdir(inputdirectory):
        mask = np.array(Image.open(inputdirectory + filename)).astype(np.uint64)
        borders = seg.find_boundaries(mask, mode='outer')
        thres_mask = np.zeros(mask.shape, dtype=np.uint8)
        thres_mask[mask == 0] = 255
        thres_mask[borders > 0] = 255
        cv2.imwrite(outputdirectory + filename + ".jpg",thres_mask)
        thres_mask = Image.fromarray(thres_mask)
        thres_mask.save(outputdirectory + filename + ".tif")
        




from utils import Dataset
import os
import numpy as np
import skimage.io 


class BowlDataset(Dataset):
    """Generates the shapes synthetic dataset. The dataset consists of simple
    shapes (triangles, squares, circles) placed randomly on a blank surface.
    The images are generated on the fly. No file access required.
    """

    def load_bowl(self, folderpaths):
        """Generate the requested number of synthetic images.
        count: number of images to generate.
        height, width: the size of the generated images.
        """
        # Add classes
        self.add_class("bowl", 1, "nuclei")

        #Image size must be dividable by 2 at least 6 times to avoid fractions when downscaling and upscaling.For example, use 256, 320, 384, 448, 512, ... etc. 

        # Add images
        for i in range(len(folderpaths)):
            self.add_image("bowl", image_id=i, path=folderpaths[i])

    def load_image(self, image_id):
        """Generate an image from the specs of the given image ID.
        Typically this function loads the image from a file, but
        in this case it generates the image on the fly from the
        specs in image_info.
        """
        info = self.image_info[image_id]
        image_path = info['path']
        image_path = os.path.join(image_path, 'images', '{}.png'.format(os.path.basename(image_path)))
        image = skimage.io.imread(image_path)
        image = image[:,:,:3]
        return image
        
    def image_reference(self, image_id):
        """Return the shapes data of the image."""
        info = self.image_info[image_id]
        if info["source"] == "bowl":
            return info["bowl"]
        else:
            super(self.__class__).image_reference(self, image_id)

    def load_mask(self, image_id):
        """Generate instance masks for shapes of the given image ID.
        """
        info = self.image_info[image_id]
        image_path = info['path']
        mask_paths = os.path.join(image_path, 'masks', '*.png')
        mask = skimage.io.imread_collection(mask_paths).concatenate()
        mask = np.rollaxis(mask,0,3)
        mask = np.clip(mask,0,1)
        class_ids = np.array([1]* mask.shape[2])
        return mask, class_ids.astype(np.int32)







import os

os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
os.environ["CUDA_VISIBLE_DEVICES"] = ""
os.environ['KERAS_BACKEND'] = 'tensorflow'

seed = 123
from keras import backend as K

import numpy as np

np.random.seed(seed)
import tensorflow as tf

tf.set_random_seed(seed) 


import random

random.seed(seed)

from PIL import Image
import skimage.transform
from skimage import img_as_ubyte

from . import model as modellib
import pandas as pd
import os

from . import my_functions as f

import time

#######################################################################################
## SET UP CONFIGURATION
from . import config

class BowlConfig(config.Config):
    """Configuration for training on the toy shapes dataset.
    Derives from the base Config class and overrides values specific
    to the toy shapes dataset.
    """
    # Give the configuration a recognizable name
    NAME = "Inference"

    IMAGE_RESIZE_MODE = "pad64"  ## tried to modfied but I am using other git clone
    ## No augmentation
    ZOOM = False
    ASPECT_RATIO = 1
    MIN_ENLARGE = 1
    IMAGE_MIN_SCALE = False  ## Not using this

    IMAGE_MIN_DIM = 512  # We scale small images up so that smallest side is 512
    IMAGE_MAX_DIM = False

    GPU_COUNT = 1
    IMAGES_PER_GPU = 1

    DETECTION_MAX_INSTANCES = 512
    DETECTION_NMS_THRESHOLD = 0.2
    DETECTION_MIN_CONFIDENCE = 0.9

    LEARNING_RATE = 0.001

    # Number of classes (including background)
    NUM_CLASSES = 1 + 1  # background + nuclei

    # Use smaller anchors because our image and objects are small
    RPN_ANCHOR_SCALES = (8, 16, 32, 64, 128)  # anchor side in pixels

    # Reduce training ROIs per image because the images are small and have
    # few objects. Aim to allow ROI sampling to pick 33% positive ROIs.
    TRAIN_ROIS_PER_IMAGE = 600

    USE_MINI_MASK = True

#######################################################################################

'''Run images through the pre-trained neural network.

Arguments:
test_path: Path where the images are stored (preprocess these using preprocess_images.py)
inputfile: Path of the comma-delimited file of images names.
outputfile: Path to write the comma-delimited run-length file to.
rescale: Set to True if rescale images before processing (saves time)
scale_factor: Multiplier to downsample images by
verbose: Verbose or not (true/false)'''
def predict_images(test_path, sample_submission, outputfilename, rescale = False, scale_factor = 2, verbose = True):
    inference_config = BowlConfig()
    ROOT_DIR = os.getcwd()
    MODEL_DIR = os.path.join(ROOT_DIR, "logs")

    output = open(outputfilename, "w")
    output.truncate()
    output.write("ImageId, EncodedPixels\n")
    output.close()

    sample_submission = pd.read_csv(sample_submission)

    if test_path[-1] != "/":
        test_path = test_path + "/"
    
    dirname = os.path.dirname(__file__)
    model_path = os.path.join(dirname, '../weights/deepretina_final.h5')

    if verbose:
        print("Loading weights from ", model_path)

    start_time = time.time()

    # Recreate the model in inference mode
    model = modellib.MaskRCNN(mode="inference",
                              config=inference_config,
                              model_dir=MODEL_DIR)
    model.load_weights(model_path, by_name=True)

    n_images = len(sample_submission.ImageId)
    for i in np.arange(n_images):
        start_time = time.time()
        image_id = sample_submission.ImageId[i]
        if verbose:
            print('Start detect', i, '  ', image_id)
        ##Set seeds for each image, just in case..
        random.seed(seed)
        np.random.seed(seed)
        tf.set_random_seed(seed)

        ## Load the image
        image_path = os.path.join(test_path, image_id, 'images', image_id + '.png')
        original_image = np.array(Image.open(image_path))

        if rescale:
            height = original_image.shape[0]
            width = original_image.shape[1]
            original_image = skimage.transform.resize(original_image, output_shape=(height // scale_factor,
                                                                                    width // scale_factor),
                                                                                    preserve_range=True)

        ####################################################################
        ## This is needed for the stage 2 image that has only one channel
        if len(original_image.shape) < 3:
            original_image = img_as_ubyte(original_image)
            original_image = np.expand_dims(original_image, 2)
            original_image = original_image[:, :, [0, 0, 0]]  # flip r and b
        ####################################################################
        original_image = original_image[:, :, :3]

        ## Make prediction for that image
        results = model.detect([original_image], verbose=0)

        ## Proccess prediction into rle
        pred_masks = results[0]['masks']
        scores_masks = results[0]['scores']
        class_ids = results[0]['class_ids']

        if len(class_ids):  ## Some objects are detected
            ImageId_batch, EncodedPixels_batch, _ = f.numpy2encoding(pred_masks, image_id, scores=scores_masks,
                                                                     dilation=True)
            f.write2csv(outputfilename, ImageId_batch, EncodedPixels_batch)
        else:
            pass

        if verbose:
            print("Completed in", time.time() - start_time)




from setuptools import setup, find_packages
import codecs
import os

VERSION = '0.0.1'
DESCRIPTION = 'Package for performing image segmentation'

# Setting up
setup(
    name="CamAgeTestThree",
    version=VERSION,
    author="Sarthak Tyagi",
    author_email="<sarthak20540@iiitd.ac.in>",
    description=DESCRIPTION,
    packages=find_packages(),
    install_requires=['keras', 'pandas', 'scipy', 'tensorflow'],
    keywords=['python', 'image', 'segmentation', 'image segmentation'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)


import numpy as np
import pandas as pd
from scipy import ndimage

def run_length_encoding(x):
    dots = np.where(x.T.flatten() == 1)[0]
    run_lengths = []
    prev = -2
    for b in dots:
        if (b>prev+1): run_lengths.extend((b + 1, 0))
        run_lengths[-1] += 1
        prev = b
    run_lengths = ' '.join([str(r) for r in run_lengths])
    return run_lengths



def remove_duplicate(mask,threshold=0.7,scores =None):
    if scores is None:
        scores = np.sum(mask,axis=(0,1)) ## Use size of nucleus as score... 
    order = np.argsort(scores)[::-1] + 1  # 1-based descending
    flat_mask = np.max(mask * np.reshape(order, [1, 1, -1]), -1)
    for i in np.arange(len(order)):
        mask[:,:,i]= mask[:,:,i]*(flat_mask == order[i])

    new_scores = np.sum(mask,axis=(0,1))
    diff_pix = scores-new_scores
    reduccion = diff_pix/scores
    # if we have reduced particle size in more than xx percent, remove particle
    ## This only has some effect during the test time augmentation merge
    mask[:,:,reduccion > threshold] = 0
    return mask
    

    
def numpy2encoding(predicts, img_name, scores=None,threshold=0.7,dilation=False):
    if dilation:
        for i in range(predicts.shape[2]): 
            predicts[:,:,i] = ndimage.binary_dilation(predicts[:,:,i])
    predicts = remove_duplicate(predicts,threshold=threshold,scores=scores)
    ImageId = []
    EncodedPixels = []
    for i in range(predicts.shape[2]): 
        rle = run_length_encoding(predicts[:,:,i])
        
        if len(rle)>0:
            ImageId.append(img_name)
            EncodedPixels.append(rle)    
    return ImageId, EncodedPixels, predicts


def write2csv(file, ImageId, EncodedPixels):
    df = pd.DataFrame({ 'ImageId' : ImageId , 'EncodedPixels' : EncodedPixels})
    with open(file, 'a', newline='') as f:
         df.to_csv(f, index=False, columns=['ImageId', 'EncodedPixels'], header=False)

    
